from abc import ABC
from typing import List, Optional, Callable

import PIL.Image


class VideoSegment(ABC):
    """
    Represents a video segment with an associated label.
    """

    @property
    def id(self):
        raise NotImplementedError()

    @property
    def label(self):
        raise NotImplementedError()

    @property
    def num_frames(self) -> int:
        raise NotImplementedError()


class VideoDataset(ABC):
    """
    A dataset interface for use with :class:`TsnDataset`. Implement this interface if you
    wish to use your dataset with TSN.

    We cannot use :class:`torch.utils.data.Dataset` because we need to yield information about
    the number of frames per video, which we can't do with the standard
    torch.utils.data.Dataset.
    """

    def __init__(
        self,
        class_count,
        segment_filter: Optional[Callable[[VideoSegment], bool]] = None,
        sample_transform: Optional[
            Callable[[List[PIL.Image.Image]], List[PIL.Image.Image]]
        ] = None,
    ) -> None:
        self.class_count = class_count
        if segment_filter is None:
            self.segment_filter = lambda _: True
        else:
            self.segment_filter = segment_filter
        if sample_transform is None:
            self.sample_transform = lambda x: x
        else:
            self.sample_transform = sample_transform

    @property
    def video_segments(self) -> List[VideoSegment]:
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def load_frames(
        self, segment: VideoSegment, idx: Optional[List[int]] = None
    ) -> List[PIL.Image.Image]:
        raise NotImplementedError()



import glob
import os
from typing import Dict, Any, List, Iterator

import pandas as pd
from gulpio.adapters import AbstractDatasetAdapter
from gulpio.utils import find_images_in_folder, resize_images

from epic_kitchens.labels import (
    VERB_CLASS_COL,
    VERB_COL,
    VIDEO_ID_COL,
    UID_COL,
    NARRATION_COL,
    PARTICIPANT_ID_COL,
)

Result = Dict[str, Any]


class EpicDatasetAdapter(AbstractDatasetAdapter):
    """Gulp Dataset Adapter for Gulping RGB frames extracted from the EPIC-KITCHENS dataset
    """

    def __init__(
        self,
        video_segment_dir: str,
        annotations_df: pd.DataFrame,
        frame_size: int = -1,
        extension: str = "jpg",
        labelled: bool = True,
    ) -> None:
        """ Gulp all action segments in  ``annotations_df`` reading the dumped frames from
        ``video_segment_dir``

        Args:
            video_segment_dir:
                Root directory containing segmented frames::

                    frame-segments/
                    ├── P01
                    │   ├── P01_01
                    │   |   ├── P01_01_0_open-door
                    │   |   |   ├── frame_0000000008.jpg
                    │   |   |   ...
                    │   |   |   ├── frame_0000000202.jpg
                    │   |   ...
                    │   |   ├── P01_01_329_put-down-plate
                    │   |   |   ├── frame_0000098424.jpg
                    │   |   |   ...
                    │   |   |   ├── frame_0000098501.jpg
                    │   ...

            annotations_df:
                DataFrame containing labels to be gulped.
            frame_size:
                Size of shortest edge of the frame, if not already this size then it will
                be resized.
            extension:
                Extension of dumped frames.
        """
        self.video_segment_dir = video_segment_dir
        self.frame_size = int(frame_size)
        self.meta_data = self._transform_annotations(annotations_df, labelled)
        self.extensions = {"jpg", "jpeg", extension}

    def iter_data(self, slice_element=None) -> Iterator[Result]:
        """Get frames and metadata corresponding to segment

        Args:
            slice_element (optional): If not specified all frames for the segment will be returned

        Yields:
            dict: dictionary with the fields

            * ``meta``: All metadata corresponding to the segment, this is the same as the data
              in the labels csv
            * ``frames``: list of :class:`PIL.Image.Image` corresponding to the frames specified
              in ``slice_element``
            * ``id``: UID corresponding to segment
        """
        slice_element = slice_element or slice(0, len(self))
        for meta in self.meta_data[slice_element]:
            clip_id = self._segment_metadata_to_clip_id(meta)
            folder = os.path.join(
                self.video_segment_dir,
                meta[PARTICIPANT_ID_COL],
                meta[VIDEO_ID_COL],
                clip_id,
            )
            paths = self._find_frames(folder)
            frames = list(resize_images(paths, self.frame_size))
            meta["frame_size"] = frames[0].shape
            meta["num_frames"] = len(frames)
            result = {"meta": meta, "frames": frames, "id": meta[UID_COL]}
            yield result

    def __len__(self):
        return len(self.meta_data)

    def _transform_annotations(
        self, annotations: pd.DataFrame, labelled: bool
    ) -> List[Dict]:
        data = []
        for i, row in annotations.iterrows():
            if labelled:
                assert not pd.isnull(
                    row[VERB_CLASS_COL]
                ), "Row at index has no verb cluster".format(i)
                assert (
                    not len(row[VERB_COL]) == 0
                ), "Row at index {} has empty verb".format(i)
            metadata = row.to_dict()
            if NARRATION_COL not in metadata:
                metadata[NARRATION_COL] = "unnarrated"
            metadata["uid"] = i
            data.append(metadata)
        return data

    def _segment_metadata_to_clip_id(self, meta):
        clip_id = "{video_id}_{uid}_{narration}".format(
            video_id=meta[VIDEO_ID_COL],
            uid=meta[UID_COL],
            narration=meta[NARRATION_COL].lower().strip().replace(" ", "-"),
        )
        return clip_id

    def _find_frames(self, folder):
        frame_paths = find_images_in_folder(folder, formats=self.extensions)
        if not len(frame_paths) > 0:
            raise MissingDataException("{} is not present".format(folder))
        return frame_paths


class EpicFlowDatasetAdapter(EpicDatasetAdapter):
    """Gulp Dataset Adapter for Gulping flow frames extracted from the EPIC-KITCHENS dataset"""

    def iter_data(self, slice_element=None):
        slice_element = slice_element or slice(0, len(self))
        for meta in self.meta_data[slice_element]:
            paths = self._find_uv_frames(meta)

            frames = {}
            for axis in "u", "v":
                frames[axis] = list(resize_images(paths[axis], self.frame_size))

            meta["frame_size"] = frames["u"][0].shape
            meta["num_frames"] = len(frames["u"])
            result = {
                "meta": meta,
                "frames": list(_intersperse(frames["u"], frames["v"])),
                "id": meta[UID_COL],
            }
            yield result

    def _find_uv_frames(self, meta):
        clip_id = self._segment_metadata_to_clip_id(meta)
        folder = {}
        paths = {}
        for axis in "u", "v":
            folder[axis] = os.path.join(
                self.video_segment_dir,
                meta[PARTICIPANT_ID_COL],
                meta[VIDEO_ID_COL],
                axis,
                clip_id,
            )
            paths[axis] = find_images_in_folder(folder[axis], formats=self.extensions)
            if not len(paths[axis]) > 0:
                raise MissingDataException("{} is not present".format(folder["u"]))
        return paths


def _intersperse(*lists):
    """
    Args:
        *lists:

    Examples:
        >>> list(_intersperse(['a', 'b']))
        ['a', 'b']
        >>> list(_intersperse(['a', 'c'], ['b', 'd']))
        ['a', 'b', 'c', 'd']
        >>> list(_intersperse(['a', 'd'], ['b', 'e'], ['c', 'f']))
        ['a', 'b', 'c', 'd', 'e', 'f']
        >>> list(_intersperse(['a', 'd', 'g'], ['b', 'e'], ['c', 'f']))
        ['a', 'b', 'c', 'd', 'e', 'f']

    """
    i = 0
    min_length = min(map(len, lists))
    total_element_count = len(lists) * min_length
    for i in range(0, total_element_count):
        list_index = i % len(lists)
        element_index = i // len(lists)
        yield lists[list_index][element_index]


class MissingDataException(Exception):
    pass



from pathlib import Path

_dataset_dir = Path(__file__).parent / "dataset"
SEGMENT_DIR = _dataset_dir / "media" / "segments"
MEDIA_DIR = _dataset_dir / "media"
ANNOTATIONS_DIR = _dataset_dir / "annotations"



import pytest

from epic_kitchens import meta


@pytest.fixture(scope="module")
def test_seen_timestamps(tmp_datadir):
    return meta.test_timestamps("seen")


@pytest.fixture(scope="module")
def test_unseen_timestamps(tmp_datadir):
    return meta.test_timestamps("unseen")


@pytest.fixture(scope="module")
def test_timestamps(test_seen_timestamps, test_unseen_timestamps):
    return meta.test_timestamps("all")


def test_seen_split_has_more_than_8000_items(test_seen_timestamps):
    assert len(test_seen_timestamps) >= 8000


def test_unseen_split_has_more_than_1000_items(test_unseen_timestamps):
    assert len(test_unseen_timestamps) >= 1000


def test_seen_split_has_uid_index(test_seen_timestamps):
    assert test_seen_timestamps.index.name == "uid"


def test_unseen_split_has_uid_index(test_unseen_timestamps):
    assert test_unseen_timestamps.index.name == "uid"


def test_all_timestamps_is_as_long_as_sum_of_both_timestamp_splits(
    test_timestamps, test_unseen_timestamps, test_seen_timestamps
):
    assert len(test_timestamps) == len(test_unseen_timestamps) + len(
        test_seen_timestamps
    )


timestamp_columns = ["video_id", "start_timestamp", "stop_timestamp", "participant_id"]


@pytest.mark.parametrize("col", timestamp_columns)
def test_seen_timestamp_has_column(col, test_seen_timestamps):
    assert col in test_seen_timestamps.columns


@pytest.mark.parametrize("col", timestamp_columns)
def test_unseen_timestamp_has_column(col, test_unseen_timestamps):
    assert col in test_unseen_timestamps.columns



"""
Loading annotations from the web is essentially the same problem regardless of what we're
loading. We want to download if the file doesn't exist, and then load it if it hasn't already been loaded
on top of this we want to version the files so we can load in different label versions in the same
session, e.g. for comparing what has changed between versions.

The process is the same:
- If the file isn't downloaded, download it to a version subfolder
- Read the file if it hasn't already been read, into an namespace divided by versions
- Potentially we also want to produce multiple data structures from each file, I'm unsure whether
  this belongs as part of this process or whether it should be separate. Consider the verb classes,
  we want to dataframe, but we also want to build an inverse lookup table from verbs to verb class.
  We could yield multiple (name, data_structure) pairs from the loading functions, but then accessing data
  structures isn't based on the filename alone, which seems a bit nasty as you have to know the name of
  the data structures returned by the loading functions to access them.

We might want to split this into a two stage process where we have this web-backed file repository
that downloads and caches the data and then another object responsible for then deriving data structures
from those files.
I'm not sure what the interface of the second object would look like though
"""
import os
import urllib.request
from ast import literal_eval

import pandas as pd
from pathlib import Path
from typing import Optional, Callable, Any, Dict
import logging

_LOG = logging.getLogger(__name__)
Loader = Callable[[Path], Any]


def _read_video_info(fp):
    df = pd.read_csv(fp, index_col="video")
    df.index.name = "video_id"
    return df


_file_loaders = {
    "EPIC_verb_classes.csv": lambda fp: pd.read_csv(
        fp, index_col="verb_id", converters={"verbs": literal_eval}
    ),
    "EPIC_noun_classes.csv": lambda fp: pd.read_csv(
        fp, index_col="noun_id", converters={"nouns": literal_eval}
    ),
    "EPIC_many_shot_actions.csv": lambda fp: set(
        pd.read_csv(
            fp, index_col="action_class", converters={"action_class": literal_eval}
        ).index.values
    ),
    "EPIC_many_shot_verbs.csv": lambda fp: set(
        pd.read_csv(fp, index_col="verb_class").index.values
    ),
    "EPIC_many_shot_nouns.csv": lambda fp: set(
        pd.read_csv(fp, index_col="noun_class").index.values
    ),
    "EPIC_descriptions.csv": lambda fp: pd.read_csv(
        fp, index_col="video_id", parse_dates={"datetime": ["date", "time"]}
    ),
    "EPIC_video_info.csv": _read_video_info,
    "EPIC_train_action_narrations.csv": lambda fp: pd.read_csv(fp),
    "EPIC_train_object_labels.csv": lambda fp: pd.read_csv(
        fp, converters={"bounding_boxes": literal_eval}
    ),
}

_filetype_loaders = {"csv": pd.read_csv, "pkl": pd.read_pickle}


class AnnotationRepository:
    def __init__(
        self, version: str = "v1.5.0", local_dir: Optional[Path] = None
    ) -> None:
        self.version = version
        base_url = "https://github.com/epic-kitchens/annotations/raw/{}/".format(
            version
        )
        if local_dir is None:
            cache_dir = Path(os.environ.get("XDG_CACHE_HOME", Path.home()))
            local_dir = cache_dir / "epic_kitchens" / version
        # local_dir.mkdir(exist_ok=True, parents=True)
        self.http_folder = HttpFolder(
            base_url, local_dir, Loaders(_file_loaders, _filetype_loaders)
        )
        self.inverse_lookups = {"verb": None, "noun": None}

    @property
    def local_dir(self):
        return self.http_folder.local_dir

    @local_dir.setter
    def local_dir(self, dir_: Path):
        self.http_folder.local_dir = dir_

    def inverse_noun_lookup(self):
        if self.inverse_lookups["noun"] is None:
            inverse_lookup = dict()
            for noun_class, row in self.noun_classes().iterrows():
                for noun in row["nouns"]:
                    inverse_lookup[noun] = noun_class
            self.inverse_lookups["noun"] = inverse_lookup
        return self.inverse_lookups["noun"]

    def inverse_verb_lookup(self):
        if self.inverse_lookups["verb"] is None:
            inverse_lookup = dict()
            for verb_class, row in self.verb_classes().iterrows():
                for verb in row["verbs"]:
                    inverse_lookup[verb] = verb_class
            self.inverse_lookups["verb"] = inverse_lookup
        return self.inverse_lookups["verb"]

    def train_action_labels(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_train_action_labels.pkl")

    def train_object_labels(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_train_object_labels.csv")

    def test_seen_timestamps(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_test_s1_timestamps.pkl")

    def test_unseen_timestamps(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_test_s2_timestamps.pkl")

    def verb_classes(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_verb_classes.csv")

    def noun_classes(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_noun_classes.csv")

    def many_shot_verbs(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_many_shot_verbs.csv")

    def many_shot_nouns(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_many_shot_nouns.csv")

    def many_shot_actions(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_many_shot_actions.csv")

    def train_action_narrations(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_train_action_narrations.csv")

    def video_descriptions(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_descriptions.csv")

    def video_info(self) -> pd.DataFrame:
        return self.http_folder.load_file("EPIC_video_info.csv")


class Loaders:
    def __init__(
        self, file_loaders: Dict[str, Loader], filetype_loaders: Dict[str, Loader]
    ) -> None:
        self.file_loaders = file_loaders
        self.filetype_loaders = filetype_loaders

    def __getitem__(self, file_name: str):
        file_type = self._get_filetype(file_name)  # drop . from extension
        if file_name in self.file_loaders:
            return self.file_loaders[file_name]
        elif file_type in self.filetype_loaders:
            return self.filetype_loaders[file_type]
        else:
            raise KeyError(
                (
                    "{} not in file_loaders, nor is a default filetype handler registered for {} "
                    "files"
                ).format(file_name, file_type)
            )

    def _get_filetype(self, file_name):
        return os.path.splitext(file_name)[1][1:]


class HttpFolder:
    def __init__(self, base_url: str, local_dir: Path, loaders: Loaders) -> None:
        self.base_url = base_url
        # TODO: Invalidate `files` cache when this is changed
        self.local_dir = local_dir
        self.loaders = loaders
        self.files = dict()  # type: Dict[str, Loader]

    def load_file(self, name: str) -> Any:
        if name not in self.files:
            file_path = self.local_dir / name
            url = self.base_url + name
            self._maybe_download(url, file_path)
            self.files[name] = self.loaders[name](file_path)
        return self.files[name]

    def _maybe_download(self, url: str, file_path: Path) -> None:
        file_path.parent.mkdir(exist_ok=True, parents=True)
        if not file_path.exists():
            urllib.request.urlretrieve(url, str(file_path))



import pandas as pd
import pytest

from epic_kitchens import meta

meta._annotation_repository.http_folder.files["EPIC_verb_classes.csv"] = pd.DataFrame(
    {
        "verb_id": [0, 1],
        "class_key": ["take", "put"],
        "verbs": [["take", "pick-up"], ["put", "put-down", "put-onto"]],
    }
)


meta._annotation_repository.http_folder.files["EPIC_noun_classes.csv"] = pd.DataFrame(
    {
        "noun_id": [0, 1, 2],
        "class_key": ["Nothing", "pan", "pan:dust"],
        "nouns": [
            ["Nothing"],
            ["pan", "pan:sauce", "saucepan"],
            ["pan:dust", "dustpan"],
        ],
    }
)


@pytest.mark.parametrize("verb_class,verb", [(0, "take"), (1, "put"), (1, "pan:sauce")])
def test_class_to_verb(verb_class, verb):
    assert meta.class_to_verb(verb_class) == verb


@pytest.mark.parametrize(
    "verb_class,verb", [(0, "take"), (0, "pick-up"), (1, "put"), (1, "put-down")]
)
def test_class_to_verb(verb_class, verb):
    assert meta.verb_to_class(verb) == verb_class


@pytest.mark.parametrize(
    "noun_class,noun", [(0, "Nothing"), (1, "pan"), (1, "pan:sauce"), (2, "pan:dust")]
)
def test_nouns_to_class(noun_class, noun):
    assert meta.noun_to_class(noun) == noun_class


@pytest.mark.parametrize(
    "noun_class,noun", [(0, "Nothing"), (1, "pan"), (2, "pan:dust")]
)
def test_class_to_nouns(noun_class, noun):
    assert meta.class_to_noun(noun_class) == noun



"""Functions for converting between frames and timestamps"""
import numpy as np

_MINUTES_TO_SECONDS = 60
_HOURS_TO_SECONDS = 60 * 60


def timestamp_to_seconds(timestamp: str) -> float:
    """ Convert a timestamp into total number of seconds

    Args:
        timestamp: formatted as ``HH:MM:SS[.FractionalPart]``

    Returns:
        ``timestamp`` converted to seconds

    Examples:
        >>> timestamp_to_seconds("00:00:00")
        0.0
        >>> timestamp_to_seconds("00:00:05")
        5.0
        >>> timestamp_to_seconds("00:00:05.5")
        5.5
        >>> timestamp_to_seconds("00:01:05.5")
        65.5
        >>> timestamp_to_seconds("01:01:05.5")
        3665.5
    """
    hours, minutes, seconds = map(float, timestamp.split(":"))
    total_seconds = hours * _HOURS_TO_SECONDS + minutes * _MINUTES_TO_SECONDS + seconds
    return total_seconds


def seconds_to_timestamp(total_seconds: float) -> str:
    """ Convert seconds into a timestamp

    Args:
        total_seconds: time in seconds

    Returns:
        timestamp representing ``total_seconds``

    Examples:
        >>> seconds_to_timestamp(1)
        '00:00:1.000'
        >>> seconds_to_timestamp(1.1)
        '00:00:1.100'
        >>> seconds_to_timestamp(60)
        '00:01:0.000'
        >>> seconds_to_timestamp(61)
        '00:01:1.000'
        >>> seconds_to_timestamp(60 * 60 + 1)
        '01:00:1.000'
        >>> seconds_to_timestamp(60 * 60  + 60 + 1)
        '01:01:1.000'
        >>> seconds_to_timestamp(1225.78500002)
        '00:20:25.785'
    """
    ss = total_seconds % 60
    mm = np.floor((total_seconds / 60) % 60)
    hh = np.floor((total_seconds / (60 * 60)))
    return "{:02.0f}:{:02.0f}:{:0.3f}".format(hh, mm, ss)


def timestamp_to_frame(timestamp: str, fps: float) -> int:
    """ Convert timestamp to frame number given the FPS of the extracted frames

    Args:
        timestamp: formatted as ``HH:MM:SS[.FractionalPart]``
        fps: frames per second

    Returns:
        frame corresponding timestamp

    Examples:
        >>> timestamp_to_frame("00:00:00", 29.97)
        1
        >>> timestamp_to_frame("00:00:01", 29.97)
        29
        >>> timestamp_to_frame("00:00:01", 59.94)
        59
        >>> timestamp_to_frame("00:01:00", 60)
        3600
        >>> timestamp_to_frame("01:00:00", 60)
        216000
    """

    total_seconds = timestamp_to_seconds(timestamp)
    if total_seconds == 0:
        return 1
    else:
        return int(np.floor(total_seconds * fps))


def flow_frame_count(rgb_frame: int, stride: int, dilation: int) -> int:
    """ Get the number of frames in a optical flow segment given the number of frames in the
    corresponding rgb segment from which the flow was extracted with parameters
    ``(stride, dilation)``

    Args:
        rgb_frame: RGB Frame number
        stride: Stride used in extracting optical flow
        dilation: Dilation used in extracting optical flow

    Returns:
       The number of optical flow frames

    Examples:
        >>> flow_frame_count(6, 1, 1)
        5
        >>> flow_frame_count(6, 2, 1)
        3
        >>> flow_frame_count(6, 1, 2)
        4
        >>> flow_frame_count(6, 2, 2)
        2
        >>> flow_frame_count(6, 3, 1)
        2
        >>> flow_frame_count(6, 1, 3)
        3
        >>> flow_frame_count(7, 1, 1)
        6
        >>> flow_frame_count(7, 2, 1)
        3
        >>> flow_frame_count(7, 1, 2)
        5
        >>> flow_frame_count(7, 2, 2)
        3
        >>> flow_frame_count(7, 3, 1)
        2
        >>> flow_frame_count(7, 1, 3)
        4
    """
    return int(np.ceil((float(rgb_frame) - dilation) / stride))



from epic_kitchens import meta


def test_default_version_is_1_5_0():
    assert meta._annotation_repository.version == "v1.5.0"


def test_changing_annotation_version():
    version = "v1.4.0"
    assert meta._annotation_repository.version != version
    meta.set_version(version)
    assert meta._annotation_repository.version == version



"""Column names present in a labels dataframe.

    Rather than accessing column names directly, we suggest you import these constants
    and use them to access the data in case the names change at any point.
"""


"""UID column name, a number uniquely identifying a particular action segment

    e.g. ``6374``"""
UID_COL = "uid"

"""Video column name, an identifier for a specific video of the form P\\d\\d_\\d\\d,\
    the first two digits are the participant ID, and the last two digits the video ID

    e.g. ``\"P03_01\"``"""
VIDEO_ID_COL = "video_id"

"""Narration column name, the original narration by the participant about the action performed

    e.g. ``\"close fridge\"``"""
NARRATION_COL = "narration"

"""Start timestamp column name, the timestamp of the start of the action segment

    e.g. ``\"00:23:43.847\"``"""
START_TS_COL = "start_timestamp"

"""Stop timestamp column name, the timestamp of the end of the action segment

    e.g. ``\"00:23:47.212\"``"""
STOP_TS_COL = "stop_timestamp"


"""Start frame column name, the frame corresponding to the starting timestamp

    e.g. ``85430``"""
START_F_COL = "start_frame"


"""Stop frame column name, the frame corresponding to the starting timestamp

    e.g. ``85643``"""
STOP_F_COL = "stop_frame"


"""Participant ID column name, the identifier corresponding to an individual

    e.g. ``85643``"""
PARTICIPANT_ID_COL = "participant_id"

"""Verb column name, the first verb extracted from the narration

    e.g. ``\"close\"``"""
VERB_COL = "verb"

"""Verb class column name, the class corresponding to the verb extracted from the narration.

    e.g. ``3``"""
VERB_CLASS_COL = "verb_class"

"""Noun column name, the first noun extracted from the narration

    e.g. ``\"fridge\"``"""
NOUN_COL = "noun"

"""Noun class column name, the class corresponding to the first noun extracted from the narration

    e.g. ``10``"""
NOUN_CLASS_COL = "noun_class"


"""Nouns column name, all nouns extracted from the narration

    e.g. ``[\"fridge\"]``"""
NOUNS_COL = "all_nouns"

"""Nouns class column name, the classes corresponding to each noun extracted from the narration

    e.g. ``[10]``"""
NOUNS_CLASS_COL = "all_noun_classes"


"""The noun class corresponding to an action without a noun, consider the narration \"stir\" where
    no object is specified."""
EMPTY_NOUN_CLASS = 0



import PIL.Image
import numpy as np
from abc import ABC
from moviepy.editor import ImageSequenceClip
from moviepy.video.io.html_tools import ipython_display
from typing import List, Union, Tuple
from IPython.display import HTML

from epic_kitchens.dataset.epic_dataset import EpicVideoDataset


def _grey_to_rgb(frames: np.ndarray) -> np.ndarray:
    """
    Convert frame(s) from gray (2D array) to RGB (3D array)

    Args:
        frames: A single frame or set of frames
    """
    if not (2 <= frames.ndim <= 3):
        raise ValueError(
            "Expected frame(s) to be a 2D array (single frame) "
            "or 3D array, but was {}D".format(frames.ndim)
        )
    n_channels = 3
    return np.repeat(np.array(frames)[..., np.newaxis], n_channels, axis=-1)


def clipify_rgb(
    frames: List[PIL.Image.Image], *, fps: float = 60.
) -> ImageSequenceClip:
    """

    Args:
        frames: A list of frames
        fps: FPS of clip

    Returns
    -------
    moviepy.editor.ImageSequenceClip
        Frames concatenated into clip



    """
    if frames == 0:
        raise ValueError("Expected at least one frame, but received none")
    frames = [np.array(frame) for frame in frames]
    segment_clip = ImageSequenceClip(frames, fps=fps)
    return segment_clip


def clipify_flow(
    frames: List[PIL.Image.Image], *, fps: float = 30.
) -> ImageSequenceClip:
    """
    Destack flow frames, join them side by side and then create a clip
    for display

    Args:
        frames:
            A list of alternating :math:`u`, :math:`v` flow frames to join into a video. Even indices should
            be :math:`u` flow frames, and odd indices, :math:`v` flow frames.

        fps: float, optional
            FPS of generated :py:class:`moviepy.editor.ImageSequenceClip`
    """
    if frames == 0:
        raise ValueError("Expected at least one frame, but received none")
    frames = _grey_to_rgb(combine_flow_uv_frames(frames))
    return ImageSequenceClip(list(frames), fps=fps)


def combine_flow_uv_frames(
    uv_frames: Union[List[PIL.Image.Image], np.ndarray],
    *,
    method="hstack",
    width_axis=2
) -> np.ndarray:
    """Destack (u, v) frames and concatenate them side by side for display purposes"""
    if isinstance(uv_frames[0], PIL.Image.Image):
        uv_frames = list(map(np.array, uv_frames))
    u_frames = np.array(uv_frames[::2])
    v_frames = np.array(uv_frames[1::2])

    return hstack_frames(u_frames, v_frames, width_axis=width_axis)


def hstack_frames(*frame_sequences: np.ndarray, width_axis: int = 2) -> np.ndarray:
    print([f.shape for f in frame_sequences])
    return np.concatenate(frame_sequences, axis=width_axis)


class Visualiser(ABC):
    def __init__(self, dataset: EpicVideoDataset) -> None:
        self.dataset = dataset

    def show(self, uid: Union[int, str], **kwargs) -> HTML:
        """
        Show the given video corresponding to ``uid`` in a HTML5 video element.

        Args:
            uid: UID of video segment
            fps (float, optional): FPS of video sequence
        """
        frames = self.dataset.load_frames(self.dataset[uid])
        clip = self._clipify_frames(frames, **kwargs)
        return ipython_display(clip)

    def _clipify_frames(
        self, frames: List[PIL.Image.Image], fps: float = 60.
    ) -> ImageSequenceClip:
        raise NotImplementedError()


class RgbVisualiser(Visualiser):
    """Visualiser for video dataset containing RGB frames"""

    def _clipify_frames(self, frames: List[PIL.Image.Image], fps: float = 60.):
        return clipify_rgb(frames, fps=fps)


class FlowVisualiser(Visualiser):
    """Visualiser for video dataset containing optical flow :math:`(u, v)` frames"""

    def _clipify_frames(self, frames: List[PIL.Image.Image], fps: float = 30):
        return clipify_flow(frames, fps=fps)



import pytest

from epic_kitchens import meta


def test_reading_verb_classes(tmp_datadir):
    verb_classes = meta.verb_classes()
    assert verb_classes["class_key"].loc[0] == "take"


def test_reading_noun_classes(tmp_datadir):
    noun_classes = meta.noun_classes()
    assert noun_classes["class_key"].loc[0] == "Nothing"


def test_reading_many_shot_verbs(tmp_datadir):
    many_shot_verbs = meta.many_shot_verbs()
    assert 0 in many_shot_verbs


def test_reading_many_shot_nouns(tmp_datadir):
    many_shot_nouns = meta.many_shot_nouns()
    assert 1 in many_shot_nouns


def test_reading_many_shot_actions(tmp_datadir):
    many_shot_actions = meta.many_shot_actions()
    assert meta.ActionClass(verb_class=0, noun_class=1) in many_shot_actions



from functools import reduce

import pytest

from epic_kitchens import meta


@pytest.fixture(scope="module")
def training_labels(tmp_datadir):
    return meta.training_labels()


@pytest.fixture(scope="module")
def training_object_labels(tmp_datadir):
    return meta.training_object_labels()


def test_training_action_labels_has_more_than_20000_items(training_labels):
    assert len(training_labels) >= 20000


def test_training_action_labels_has_uid_index(training_labels):
    assert training_labels.index.name == "uid"


@pytest.mark.parametrize(
    "col",
    [
        "video_id",
        "narration",
        "start_timestamp",
        "stop_timestamp",
        "participant_id",
        "verb",
        "noun",
        "verb_class",
        "noun_class",
        "all_nouns",
        "all_noun_classes",
    ],
)
def test_training_action_labels_has_column(col, training_labels):
    assert col in training_labels.columns


@pytest.mark.parametrize(
    "col",
    ["noun", "noun_class", "frame", "participant_id", "video_id", "bounding_boxes"],
)
def test_training_object_labels_has_column(col, training_object_labels):
    assert col in training_object_labels.columns


def test_training_object_labels_has_over_350_000_rows(training_object_labels):
    assert len(training_object_labels) >= 350000


def test_training_object_labels_has_list_of_bounding_boxes_per_row(
    training_object_labels
):
    assert (
        training_object_labels["bounding_boxes"]
        .apply(lambda bbs: isinstance(bbs, list))
        .all()
    )


def test_training_object_labels_has_4_tuple_bounding_boxes(training_object_labels):
    def all_fours(all_fours_so_far: bool, bounding_boxes):
        return all_fours_so_far and len(bounding_boxes) == 4

    assert (
        training_object_labels["bounding_boxes"]
        .apply(lambda bbs: reduce(all_fours, bbs, True))
        .all()
    )



import pandas as pd
from collections import namedtuple

from epic_kitchens.internal.loading import AnnotationRepository

from pathlib import Path
from typing import Union, Set
from logging import getLogger

_LOG = getLogger(__name__)

_annotation_repositories = {"v1.5.0": AnnotationRepository()}
_annotation_repository = _annotation_repositories["v1.5.0"]

ActionClass = namedtuple("ActionClass", ["verb_class", "noun_class"])
Action = namedtuple("Action", ["verb", "noun"])


def set_version(version: str):
    global _annotation_repository, _annotation_repositories
    if version not in _annotation_repositories:
        _annotation_repositories[version] = AnnotationRepository(
            version=version, local_dir=_annotation_repository.local_dir
        )
    _annotation_repository = _annotation_repositories[version]


def set_datadir(dir_: Union[str, Path]):
    """Set download directory

    Args:
        dir_:
            Path to directory in which to store all downloaded metadata files
    """
    _annotation_repository.local_dir = Path(dir_)
    _LOG.info("Setting data directory to {}".format(dir_))


def get_datadir() -> Path:
    """
    Returns:
        Directory under which any downloaded files are stored, defaults to current working directory
    """
    return _annotation_repository.local_dir


def verb_to_class(verb: str) -> int:
    """
    Args:
        verb:
            A noun from a narration

    Returns:
        The corresponding numeric class of the verb if it exists

    Raises:
        IndexError:
            If the verb doesn't belong to any of the verb classes
    """
    return _annotation_repository.inverse_verb_lookup()[verb]


def noun_to_class(noun: str) -> int:
    """
    Args:
        noun:
            A noun from a narration

    Returns:
        The corresponding numeric class of the noun if it exists

    Raises:
        IndexError:
            If the noun doesn't belong to any of the noun classes
    """
    return _annotation_repository.inverse_noun_lookup()[noun]


def class_to_verb(cls: int) -> str:
    """
    Args:
        cls: numeric verb class

    Returns:
        Canonical verb representing the class

    Raises:
        IndexError:
            if ``cls`` is an invalid verb class
    """
    return _annotation_repository.verb_classes()["class_key"].loc[cls]


def class_to_noun(cls: int) -> str:
    """
    Args:
        cls: numeric noun class

    Returns:
        Canonical noun representing the class

    Raises:
        IndexError:
            if ``cls`` is an invalid noun class

    """
    return _annotation_repository.noun_classes()["class_key"].loc[cls]


def noun_classes() -> pd.DataFrame:
    """
    Get dataframe containing the mapping between numeric noun classes, the canonical noun of that class
    and nouns clustered into the class.

    Returns:
        Dataframe with the columns:

        .. include:: meta/noun_classes.rst
    """
    return _annotation_repository.noun_classes().copy()


def verb_classes() -> pd.DataFrame:
    """
    Get dataframe containing the mapping between numeric verb classes, the canonical verb of that class
    and verbs clustered into the class.

    Returns:
        Dataframe with the columns

        .. include:: meta/verb_classes.rst
    """
    return _annotation_repository.verb_classes().copy()


def many_shot_verbs() -> Set[int]:
    """
    Returns:
        The set of verb classes that are many shot (appear more than 100 times in training).
    """
    return set(_annotation_repository.many_shot_verbs())


def many_shot_nouns() -> Set[int]:
    """
    Returns:
        The set of noun classes that are many shot (appear more than 100 times in training).
    """
    return set(_annotation_repository.many_shot_nouns())


def many_shot_actions() -> Set[ActionClass]:
    """
    Returns:
        The set of actions classes that are many shot (verb_class appears more than 100 times
        in training, noun_class appears more than 100 times in training, and the action appears
        at least once in training).

    """
    return set(_annotation_repository.many_shot_actions())


def is_many_shot_action(action_class: ActionClass) -> bool:
    """
    Args:
        action_class:
            ``(verb_class, noun_class)`` tuple

    Returns:
        Whether action_class is many shot or not
    """
    return action_class in _annotation_repository.many_shot_actions()


def is_many_shot_verb(verb_class: int) -> bool:
    """
    Args:
        verb_class: numeric verb class

    Returns:
        Whether verb_class is many shot or not
    """
    return verb_class in _annotation_repository.many_shot_verbs()


def is_many_shot_noun(noun_class: int) -> bool:
    """
    Args:
        noun_class: numeric noun class

    Returns:
        Whether noun class is many shot or not

    """
    return noun_class in _annotation_repository.many_shot_nouns()


def training_narrations() -> pd.DataFrame:
    """
    Returns:
        Dataframe with the columns

        .. include:: meta/train_action_narrations.rst
    """
    return _annotation_repository.train_action_narrations()


def training_labels() -> pd.DataFrame:
    """
    Returns:
        Dataframe with the columns

        .. include:: meta/train_action_labels.rst
    """
    return _annotation_repository.train_action_labels().copy()


def training_object_labels() -> pd.DataFrame:
    """
    Returns:
        Dataframe with the columns

        .. include:: meta/train_object_labels.rst
    """
    return _annotation_repository.train_object_labels().copy()


def test_timestamps(split: str) -> pd.DataFrame:
    """
    Args:
        split: 'seen', 'unseen', or 'all' (loads both with a 'split'

    Returns:
        Dataframe with the columns

        .. include:: meta/test_timestamps.rst
    """
    if split == "all":
        return pd.concat(
            [
                _annotation_repository.test_seen_timestamps(),
                _annotation_repository.test_unseen_timestamps(),
            ]
        )

    if split == "seen":
        return _annotation_repository.test_seen_timestamps().copy()
    elif split == "unseen":
        return _annotation_repository.test_unseen_timestamps()
    raise ValueError("Unknown split '{}' expected 'seen' or 'unseen'".format(split))


def video_descriptions() -> pd.DataFrame:
    """
    Returns:
        High level description of the task trying to be accomplished in a video

        .. include:: meta/video_descriptions.rst
    """
    return _annotation_repository.video_descriptions().copy()


def video_info() -> pd.DataFrame:
    """
    Returns:
        Technical information stating the resolution, duration and FPS of each video.

        .. include:: meta/video_info.rst
    """
    return _annotation_repository.video_info().copy()



import cv2
from skimage.measure import compare_ssim as ssim

from epic_kitchens.gulp.__main__ import main, parser
from gulpio.dataset import GulpDirectory
import pandas as pd
import numpy as np

from tests import SEGMENT_DIR, ANNOTATIONS_DIR


def test_gulping_labelled_rgb_segments(tmpdir):
    gulp_dir_path = tmpdir.mkdir("rgb_labelled_gulped")
    segment_dir = SEGMENT_DIR / "rgb"
    annotations_path = ANNOTATIONS_DIR / "EPIC_train_action_labels.csv"

    gulp(segment_dir, gulp_dir_path, annotations_path, "rgb", True)
    gulp_dir = GulpDirectory(str(gulp_dir_path))

    assert_number_of_segments(gulp_dir, 11)
    annotations = pd.read_csv(annotations_path, index_col="uid")
    for annotation in annotations.itertuples():
        metadata = gulp_dir.merged_meta_dict[str(annotation.Index)]["meta_data"][0]
        assert metadata["verb_class"] == annotation.verb_class
        assert metadata["noun_class"] == annotation.noun_class

    assert_gulped_rgb_frames_similar_to_on_disk(
        gulp_dir, annotations, segment_dir, 1, max_discrepancy=10
    )


def test_gulping_unlabelled_rgb_segments(tmpdir):
    gulp_dir_path = tmpdir.mkdir("rgb_labelled_gulped")
    segment_dir = SEGMENT_DIR / "rgb"
    annotations_path = ANNOTATIONS_DIR / "EPIC_test_s1_timestamps.csv"

    gulp(segment_dir, gulp_dir_path, annotations_path, "rgb", False)
    gulp_dir = GulpDirectory(str(gulp_dir_path))

    assert_number_of_segments(gulp_dir, 7)
    annotations = pd.read_csv(annotations_path, index_col="uid")
    for annotation in annotations.itertuples():
        metadata = gulp_dir.merged_meta_dict[str(annotation.Index)]["meta_data"][0]
        assert metadata["uid"] == annotation.Index


def test_gulping_labelled_flow_segments(tmpdir):
    gulp_dir_path = tmpdir.mkdir("flow_labelled_gulped")
    segment_dir = SEGMENT_DIR / "flow"
    annotations_path = ANNOTATIONS_DIR / "EPIC_train_action_labels.csv"

    gulp(segment_dir, gulp_dir_path, annotations_path, "flow", True)
    gulp_dir = GulpDirectory(str(gulp_dir_path))

    assert_number_of_segments(gulp_dir, 11)
    annotations = pd.read_csv(annotations_path, index_col="uid")
    for annotation in annotations.itertuples():
        metadata = gulp_dir.merged_meta_dict[str(annotation.Index)]["meta_data"][0]
        assert metadata["verb_class"] == annotation.verb_class
        assert metadata["noun_class"] == annotation.noun_class

    assert_gulped_flow_frames_similar_to_on_disk(gulp_dir, annotations, segment_dir, 0)


def test_gulping_unlabelled_flow_segments(tmpdir):
    gulp_dir_path = tmpdir.mkdir("flow_unlabelled_gulped")
    segment_dir = SEGMENT_DIR / "flow"
    annotations_path = ANNOTATIONS_DIR / "EPIC_test_s1_timestamps.csv"

    gulp(segment_dir, gulp_dir_path, annotations_path, "flow", False)
    gulp_dir = GulpDirectory(str(gulp_dir_path))

    assert_number_of_segments(gulp_dir, 7)
    annotations = pd.read_csv(annotations_path, index_col="uid")
    for annotation in annotations.itertuples():
        metadata = gulp_dir.merged_meta_dict[str(annotation.Index)]["meta_data"][0]
        assert metadata["uid"] == annotation.Index


def assert_number_of_segments(gulp_dir, number_of_segments):
    segment_count = number_of_segments
    assert len(gulp_dir.merged_meta_dict) == segment_count


def assert_gulped_flow_frames_similar_to_on_disk(
    gulp_dir, annotations, segment_dir, uid, min_ssim=0.95
):
    """
    Assert that the first 2 gulped frames for a given UID are close to those on disk.
    ``max_discrepancy`` controls how much of a difference per pixel is tolerable, set it to a value in
    [0, 255]
    """
    for axis in "u", "v":
        segment_path = get_segment_path(segment_dir, annotations, axis, uid)
        frame_paths = [segment_path / ("frame_%010d.jpg" % i) for i in (1, 2)]
        u_frames = read_images(frame_paths)
        gulp_frames, _ = gulp_dir[uid, slice(0, 4)]
        gulp_u_frames = gulp_frames[::2]
        gulp_v_frames = gulp_frames[1::2]
        gulp_frames = gulp_u_frames if axis == "u" else gulp_v_frames
        for frame, gulp_frame in zip(u_frames, gulp_frames):
            assert frame.shape == gulp_frame.shape

            computed_ssim = ssim(gulp_frame, frame, data_range=255.0)
            assert computed_ssim >= min_ssim


def assert_gulped_rgb_frames_similar_to_on_disk(
    gulp_dir, annotations, segment_dir, uid, max_discrepancy=1
):
    """
    Assert that the first 2 gulped frames for a given UID are close to those on disk.
    ``max_discrepancy`` controls how much of a difference per pixel is tolerable, set it to a value in [0, 255]
    """
    segment_path = get_segment_path(segment_dir, annotations, None, uid)
    frame_paths = [segment_path / ("frame_%010d.jpg" % i) for i in range(1, 3)]
    frames = read_images(frame_paths)
    gulp_frames, _ = gulp_dir[uid, slice(0, 2)]
    for i, (frame, gulp_frame) in enumerate(zip(frames, gulp_frames)):
        assert frame.shape == gulp_frame.shape

        # We let there be a maximum discrepancy of 1 per pixel throughout the image
        max_discrepancy = np.prod(frame.shape) * max_discrepancy
        discrepancy = np.sum(np.abs(frame - gulp_frame))
        assert discrepancy <= max_discrepancy


def get_segment_path(root_segment_path, annotations, axis, uid):
    annotation = annotations.loc[uid]
    participant_id, video_id, narration = annotation[
        ["participant_id", "video_id", "narration"]
    ]
    segment_name = "{}_{}_{}".format(video_id, uid, narration.strip().replace(" ", "-"))
    pre_axis_path = root_segment_path / participant_id / video_id
    if axis:
        return pre_axis_path / axis / segment_name
    else:
        return pre_axis_path / segment_name


def read_images(paths):
    images = []
    for path in paths:
        image = cv2.imread(str(path), cv2.IMREAD_ANYCOLOR)
        if image.ndim == 3:
            image = bgr_to_rgb(image)
        images.append(image)
    return images


def bgr_to_rgb(image):
    return image[..., ::-1]


def gulp(segment_root_dir, gulp_dir, label_path, modality, labelled):
    main(
        parser.parse_args(
            [str(segment_root_dir), str(gulp_dir), str(label_path), modality]
            + (["--unlabelled"] if not labelled else [])
        )
    )



import copy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, Iterable

import PIL.Image
from gulpio import GulpDirectory

from epic_kitchens.labels import VERB_CLASS_COL, NOUN_CLASS_COL, UID_COL
from epic_kitchens.dataset.video_dataset import VideoDataset, VideoSegment


SegmentFilter = Callable[[VideoSegment], bool]
ClassGetter = Callable[[Dict[str, Any]], Any]
VideoTransform = Callable[[List[PIL.Image.Image]], List[PIL.Image.Image]]


def _verb_class_getter(metadata):
    return int(metadata[VERB_CLASS_COL])


def _noun_class_getter(metadata):
    return int(metadata[NOUN_CLASS_COL])


_class_getters = {
    "verb": _verb_class_getter,
    "noun": _noun_class_getter,
    "verb+noun": lambda metadata: {
        "verb": _verb_class_getter(metadata),
        "noun": _noun_class_getter(metadata),
    },
    None: lambda meta: None,
}

_verb_class_count = 125
_noun_class_count = 353
_class_count = {
    "verb": _verb_class_count,
    "noun": _noun_class_count,
    "verb+noun": (_verb_class_count, _noun_class_count),
    None: 0,
}


class GulpVideoSegment(VideoSegment):
    """SegmentRecord for a video segment stored in a gulp file.

    Assumes that the video segment has the following metadata in the gulp file:
      - id
      - num_frames
    """

    def __init__(
        self,
        gulp_metadata_dict: Dict[str, Any],
        class_getter: Callable[[Dict[str, Any]], Any],
    ) -> None:
        self.metadata = gulp_metadata_dict
        self.class_getter = class_getter
        self.gulp_index = gulp_metadata_dict[UID_COL]

    @property
    def id(self) -> str:
        """ID of video segment"""
        return self.gulp_index

    @property
    def label(self) -> Any:
        cls = self.class_getter(self.metadata)
        # WARNING: this type check should be removed once we regulp our data
        # so that classes are ints in the metadata json
        if isinstance(cls, float):
            return int(cls)
        else:
            return cls

    @property
    def num_frames(self) -> int:
        """Number of video frames"""
        return self.metadata["num_frames"]

    def __getitem__(self, item):
        return self.metadata[item]

    def __getattr__(self, item):
        return self.metadata[item]

    def __str__(self):
        return "GulpVideoSegment[label={label}, num_frames={num_frames}]".format(
            label=self.label, num_frames=self.num_frames
        )

    def __repr__(self):
        return "GulpVideoSegment({metadata}, {class_getter})".format(
            metadata=repr(self.metadata), class_getter=repr(self.class_getter)
        )


class EpicVideoDataset(VideoDataset):
    """VideoDataset for gulped RGB frames"""

    def __init__(
        self,
        gulp_path: Union[Path, str],
        class_type: str,
        *,
        with_metadata: bool = False,
        class_getter: Optional[ClassGetter] = None,
        segment_filter: Optional[SegmentFilter] = None,
        sample_transform: Optional[VideoTransform] = None
    ) -> None:
        """
        Args:
            gulp_path: Path to gulp directory containing the gulped EPIC RGB or flow frames

            class_type: One of verb, noun, verb+noun, None, determines what label the segment
                returns. ``None`` should be used for loading test datasets.

            with_metadata: When True the segments will yield a tuple (metadata, class) where the
                class is defined by the class getter and the metadata is the raw dictionary stored
                in the gulp file.

            class_getter: Optionally provide a callable that takes in the gulp dict representing the
                segment from which you should return the class you wish the segment to have.

            segment_filter: Optionally provide a callable that takes a segment and returns True if
                you want to keep the segment in the dataset, or False if you wish to exclude it.

            sample_transform: Optionally provide a sample transform function which takes a list of
                PIL images and transforms each of them. This is applied on the frames just before
                returning from :meth:`load_frames`.
        """
        super().__init__(
            _class_count[class_type],
            segment_filter=segment_filter,
            sample_transform=sample_transform,
        )
        if isinstance(gulp_path, str):
            gulp_path = Path(gulp_path)
        assert gulp_path.exists(), "Could not find the path {}".format(gulp_path)
        self.gulp_dir = GulpDirectory(str(gulp_path))
        if class_getter is None:
            class_getter = _class_getters[class_type]
        if with_metadata:
            original_getter = copy.copy(class_getter)
            class_getter = lambda metadata: (metadata, original_getter(metadata))
        self._video_segments = self._read_segments(
            self.gulp_dir.merged_meta_dict, class_getter
        )

    @property
    def video_segments(self) -> List[VideoSegment]:
        """
        List of video segments that are present in the dataset. The describe the start and stop
        times of the clip and its class.
        """
        return list(self._video_segments.values())

    def load_frames(
        self, segment: VideoSegment, indices: Optional[Iterable[int]] = None
    ) -> List[PIL.Image.Image]:
        """
        Load frame(s) from gulp directory.

        Args:
            segment: Video segment to load
            indices: Frames indices to read

        Returns:
            Frames indexed by ``indices`` from the ``segment``.

        """
        if indices is None:
            indices = range(0, segment.num_frames)
        selected_frames = []  # type: List[PIL.Image.Image]
        for i in indices:
            # Without passing a slice to the gulp directory index we load ALL the frames
            # so we create a slice with a single element -- that way we only read a single frame
            # from the gulp chunk, and not the whole chunk.
            # Here we also apply the sample transform to the loaded frames
            frames = self._sample_video_at_index(segment, i)
            frames = self.sample_transform(frames)
            selected_frames.extend(frames)
        return selected_frames

    def __len__(self):
        return len(self.video_segments)

    def __getitem__(self, id):
        return self._video_segments[id]

    def __contains__(self, id):
        return id in self._video_segments

    def _read_segments(
        self, gulp_dir_meta_dict, class_getter: Callable[[Dict[str, Any]], Any]
    ) -> Dict[str, VideoSegment]:
        segments = dict()  # type: Dict[str, VideoSegment]
        for video_id in gulp_dir_meta_dict:
            segment = GulpVideoSegment(
                gulp_dir_meta_dict[video_id]["meta_data"][0], class_getter
            )
            if self.segment_filter(segment):
                segments[segment.id] = segment
        return segments

    def _sample_video_at_index(
        self, record: VideoSegment, index: int
    ) -> List[PIL.Image.Image]:
        single_frame_slice = slice(index, index + 1)
        numpy_frame = self.gulp_dir[record.id, single_frame_slice][0][0]
        return [PIL.Image.fromarray(numpy_frame).convert("RGB")]


class EpicVideoFlowDataset(EpicVideoDataset):
    """VideoDataset for loading gulped flow. The loader assumes that flow :math:`u`, :math:`v`
    frames are stored alternately in a flat manner: :math:`[u_0, v_0, u_1, v_1, \ldots, u_n, v_n]`

    """

    def _sample_video_at_index(
        self, record: VideoSegment, index: int
    ) -> List[PIL.Image.Image]:
        # Flow pairs are stored in a contiguous manner in the gulp chunk:
        # [u_1, v_1, u_2, v_2, ..., u_n, v_n]
        # so we have to convert our desired frame index i to the gulp
        # indices j by j = (i * 2, (i + 1) * 2)
        flow_pair_slice = slice(index * 2, (index + 1) * 2)
        numpy_frames = self.gulp_dir[record.id, flow_pair_slice][0]
        frames = [
            PIL.Image.fromarray(numpy_frame).convert("L")
            for numpy_frame in numpy_frames
        ]
        return frames



__title__ = "epic-kitchens"
__description__ = "EPIC Kitchens action dataset support library"
__version__ = "1.6.3"
__author__ = "EPIC KITCHENS"
__author_email__ = "uob-epic-kitchens2018@bristol.ac.uk"
__license__ = "MIT"
__copyright__ = "Copyright 2018 EPIC Kitchens"



import os
import pathlib

import pytest
from pathlib import Path
from unittest.mock import MagicMock
import urllib.request

from epic_kitchens.internal.loading import Loaders, HttpFolder, AnnotationRepository

base_url = "https://github.com/epic-kitchens/annotations/raw/v1.5.0/"
file_name = "EPIC_video_info.csv"
loaders = {file_name: MagicMock()}


@pytest.fixture(scope="function")
def http_folder(tmpdir_factory):
    local_dir = Path(str(tmpdir_factory.mktemp("v1.5.0")))
    folder = HttpFolder(base_url, local_dir, loaders)
    yield folder
    for loader in loaders.values():
        loader.reset_mock()


@pytest.fixture(scope="function")
def dummy_path_mkdir(monkeypatch):
    with monkeypatch.context() as ctx:
        ctx.setattr(Path, "mkdir", lambda self, **kwargs: None)
        yield


@pytest.fixture(scope="function")
def dummy_urlretrieve(monkeypatch):
    with monkeypatch.context() as ctx:
        ctx.setattr(urllib.request, "urlretrieve", MagicMock())
        yield


@pytest.mark.usefixtures("dummy_path_mkdir", "dummy_urlretrieve")
class TestHttpFolder:
    def test_load_file_downloads_file_if_it_doesnt_exist(
        self, monkeypatch, http_folder
    ):
        with monkeypatch.context() as ctx:
            urlretrieve = MagicMock()
            ctx.setattr(urllib.request, "urlretrieve", urlretrieve)
            http_folder.load_file(file_name)

        urlretrieve.assert_called_once_with(
            base_url + file_name, str(http_folder.local_dir / file_name)
        )

    def test_load_file_uses_loader_for_file(self, monkeypatch, http_folder):
        http_folder.load_file(file_name)

        loaders[file_name].assert_called_once_with(http_folder.local_dir / file_name)

    def test_load_file_caches_file_load(self, monkeypatch, http_folder):
        with monkeypatch.context() as ctx:
            http_folder.load_file(file_name)
            http_folder.load_file(file_name)

        loaders[file_name].assert_called_once_with(http_folder.local_dir / file_name)

    def test_load_file_creates_parent_dir_if_it_doesnt_exist(
        self, monkeypatch, http_folder
    ):
        mkdir_mock = MagicMock()
        with monkeypatch.context() as ctx:
            ctx.setattr(Path, "mkdir", mkdir_mock)
            http_folder.load_file(file_name)
        mkdir_mock.assert_called_once_with(exist_ok=True, parents=True)


class TestLoaders:
    def test_loader_defaults_to_filetype_loader_for_unknown_file(self):
        filetype_loaders = {"pkl": MagicMock()}
        loaders = Loaders(dict(), filetype_loaders)
        assert loaders["unknown_file.pkl"] == filetype_loaders["pkl"]

    def test_loader_uses_file_specific_loader_for_known_file(self):
        filetype_loaders = {"pkl": MagicMock()}
        filename = "EPIC_train_action_labels.pkl"
        file_loaders = {filename: MagicMock()}
        loaders = Loaders(file_loaders, filetype_loaders)

        assert loaders[filename] == file_loaders[filename]


class TestAnnotationRepository:
    def test_defaults_to_user_cache_dir_for_storage_location(self):
        repo = AnnotationRepository()
        assert repo.http_folder.local_dir == Path.home() / "epic_kitchens" / "v1.5.0"

    def test_uses_version_string_in_storage_location(self):
        version = "v0.0.1"
        repo = AnnotationRepository(version=version)
        assert version in str(repo.http_folder.local_dir)

    def test_uses_version_string_in_url(self):
        version = "v0.0.1"
        repo = AnnotationRepository(version=version)
        assert repo.http_folder.base_url.endswith(version + "/")

    def test_uses_XDG_CACHE_HOME_if_set_for_storage_location(self, monkeypatch):
        cache_dir = "/home/will/.mycache"
        with monkeypatch.context() as ctx:
            ctx.setattr(os.environ, "get", lambda name, default: cache_dir)
            repo = AnnotationRepository()
        assert cache_dir in str(repo.http_folder.local_dir)



""" Dataset Adapters for GulpIO.

This module contains two adapters for 'gulping' both RGB and flow frames
which can then be used with the :class:`EpicVideoDataset` classes.
"""

from . import adapter



from pathlib import Path

from epic_kitchens.preprocessing.split_segments import main, parser
from tests import ANNOTATIONS_DIR, MEDIA_DIR


def test_splitting_labelled_rgb(tmpdir):
    participant_id = "P01"
    video_id = "P01_01"
    modality = "rgb"
    p01_01_dir = get_segment_dir(tmpdir, participant_id, video_id)
    frame_dir = get_frame_dir(modality, participant_id, video_id)
    annotations_path = ANNOTATIONS_DIR / "EPIC_train_action_labels.csv"

    split_actions(video_id, frame_dir, p01_01_dir, annotations_path, modality)

    assert (p01_01_dir / "P01_01_0_open-door" / "frame_0000000001.jpg").exists()
    assert (p01_01_dir / "P01_01_0_open-door" / "frame_0000000194.jpg").exists()
    assert (p01_01_dir / "P01_01_10_open-drawer" / "frame_0000000001.jpg").exists()
    assert (p01_01_dir / "P01_01_10_open-drawer" / "frame_0000000074.jpg").exists()


def test_splitting_unlabelled_rgb(tmpdir):
    participant_id = "P01"
    video_id = "P01_11"
    modality = "rgb"
    p01_11_dir = get_segment_dir(tmpdir, participant_id, video_id)
    frame_dir = get_frame_dir(modality, participant_id, video_id)
    annotations_path = ANNOTATIONS_DIR / "EPIC_test_s1_timestamps.csv"

    split_actions(video_id, frame_dir, p01_11_dir, annotations_path, modality)

    assert (p01_11_dir / "P01_11_1924_unnarrated").exists()
    assert (p01_11_dir / "P01_11_1924_unnarrated" / "frame_0000000001.jpg").exists()
    assert (p01_11_dir / "P01_11_1924_unnarrated" / "frame_0000000112.jpg").exists()

    assert (p01_11_dir / "P01_11_1930_unnarrated").exists()
    assert (p01_11_dir / "P01_11_1930_unnarrated" / "frame_0000000001.jpg").exists()
    assert (p01_11_dir / "P01_11_1930_unnarrated" / "frame_0000000407.jpg").exists()


def test_splitting_labelled_flow(tmpdir):
    participant_id = "P01"
    video_id = "P01_01"
    modality = "flow"
    p01_01_dir = get_segment_dir(tmpdir, participant_id, video_id)
    frame_dir = get_frame_dir(modality, participant_id, video_id)
    annotations_path = ANNOTATIONS_DIR / "EPIC_train_action_labels.csv"

    split_actions(video_id, frame_dir, p01_01_dir, annotations_path, modality)

    for axis in "u", "v":
        assert (
            p01_01_dir / axis / "P01_01_0_open-door" / "frame_0000000001.jpg"
        ).exists()
        assert (
            p01_01_dir / axis / "P01_01_0_open-door" / "frame_0000000097.jpg"
        ).exists()
        assert (
            p01_01_dir / axis / "P01_01_10_open-drawer" / "frame_0000000001.jpg"
        ).exists()
        assert (
            p01_01_dir / axis / "P01_01_10_open-drawer" / "frame_0000000037.jpg"
        ).exists()


def test_splitting_unlabelled_flow(tmpdir):
    participant_id = "P01"
    video_id = "P01_11"
    modality = "flow"
    p01_11_dir = get_segment_dir(tmpdir, participant_id, video_id)
    frame_dir = get_frame_dir(modality, participant_id, video_id)
    annotations_path = ANNOTATIONS_DIR / "EPIC_test_s1_timestamps.csv"

    split_actions(video_id, frame_dir, p01_11_dir, annotations_path, modality)

    for axis in "u", "v":
        assert (
            p01_11_dir / axis / "P01_11_1924_unnarrated" / "frame_0000000001.jpg"
        ).exists()
        assert (
            p01_11_dir / axis / "P01_11_1924_unnarrated" / "frame_0000000056.jpg"
        ).exists()

        assert (p01_11_dir / axis / "P01_11_1930_unnarrated").exists()
        assert (
            p01_11_dir / axis / "P01_11_1930_unnarrated" / "frame_0000000001.jpg"
        ).exists()
        assert (
            p01_11_dir / axis / "P01_11_1930_unnarrated" / "frame_0000000204.jpg"
        ).exists()


def split_actions(video_id, frame_dir, segment_dir, annotations_path, modality):
    main(
        parser.parse_args(
            [
                video_id,
                str(frame_dir),
                str(segment_dir),
                str(annotations_path),
                modality,
            ]
        )
    )


def get_frame_dir(modality, participant_id, video_id):
    return MEDIA_DIR / modality / participant_id / video_id


def get_segment_dir(tmpdir, participant_id, video_id):
    segments_dir = Path(str(tmpdir.mkdir("segments")))
    video_segments_dir = segments_dir / participant_id / video_id
    video_segments_dir.mkdir(parents=True, exist_ok=False)
    return video_segments_dir



import pytest
import epic_kitchens.meta


@pytest.fixture(scope="module")
def training_narrations(tmp_datadir):
    return epic_kitchens.meta.training_narrations()


def test_training_narrations_has_at_least_20_000_entries(training_narrations):
    assert len(training_narrations) >= 20000


@pytest.mark.parametrize(
    "col",
    ["participant_id", "video_id", "start_timestamp", "stop_timestamp", "narration"],
)
def test_training_narrations_has_columns(col, training_narrations):
    assert col in training_narrations.columns



import pytest

import epic_kitchens.meta


@pytest.fixture(scope="module")
def video_info(tmp_datadir):
    return epic_kitchens.meta.video_info()


def test_video_info_has_video_id_index(video_info):
    assert video_info.index.name == "video_id"


def test_video_info_has_at_least_432_entries(video_info):
    assert len(video_info) >= 432


@pytest.mark.parametrize("col", ["resolution", "duration", "fps"])
def test_video_info_has_columns(col, video_info):
    assert col in video_info.columns



import urllib.request
from pathlib import Path
from unittest.mock import Mock, MagicMock

from epic_kitchens.internal.utils import before, maybe_download


URL_ROOT = "https://raw.githubusercontent.com/epic-kitchens/annotations/master/"


def test_before_decorator():
    fn = lambda x: "blah"
    before_fn = Mock()
    decorator = before(before_fn)
    wrapped_fn = decorator(fn)

    assert wrapped_fn("x") == "blah"
    before_fn.assert_called_once_with()


def test_maybe_download_when_file_doesnt_exist(monkeypatch):
    monkeypatch.setattr(Path, "exists", lambda self: False)
    urlretrieve = MagicMock()
    monkeypatch.setattr(urllib.request, "urlretrieve", urlretrieve)

    url = URL_ROOT + "EPIC_verb_classes.csv"
    maybe_download(url, "file")

    urlretrieve.assert_called_once_with(url, "file")


def test_maybe_download_when_file_does_exist(monkeypatch):
    monkeypatch.setattr(Path, "exists", lambda self: True)
    urlretrieve = MagicMock()
    monkeypatch.setattr(urllib.request, "urlretrieve", urlretrieve)

    url = URL_ROOT + "EPIC_verb_classes.csv"
    maybe_download(url, "file")

    urlretrieve.assert_not_called()



import logging
import urllib.request
from pathlib import Path

import functools
from typing import Union


def before(before_fn):
    """
    Build a decorator function that will apply ``before_fn`` before calling the decorated function.
    """

    def decorator(fn):
        def wrapped_fn(*args, **kwargs):
            before_fn()
            return fn(*args, **kwargs)

        return functools.update_wrapper(wrapped_fn, fn)

    return decorator


def maybe_download(url: str, filepath: Union[Path, str]):
    filepath = Path(filepath)
    if not filepath.exists():
        _LOG.info("Downloading {} to {}".format(url, filepath))
        urllib.request.urlretrieve(url, str(filepath))


_LOG = logging.getLogger(__name__)



"""Program for building GulpIO directory of frames for training
See :ref:`cli_tools_gulp_ingestor` for usage details """
from pathlib import Path
import argparse
import pandas as pd
from gulpio import GulpIngestor

from . import adapter


parser = argparse.ArgumentParser(
    "Gulp the EPIC dataset allowing for faster read times during training."
)
parser.add_argument(
    "in_folder",
    type=Path,
    help="Directory where subdirectory is a segment name containing frames for that segment.",
)
parser.add_argument(
    "out_folder", type=Path, help="Directory to store the gulped files."
)
parser.add_argument(
    "labels",
    type=Path,
    help="Path to the pickle or CSV file which contains the meta information about the dataset.",
)
parser.add_argument("modality", choices=["flow", "rgb"])
parser.add_argument(
    "--extension",
    type=str,
    default="jpg",
    help="Which file extension the frames are saved as.",
)
parser.add_argument("--frame-size", type=int, default=-1, help="Size of frames.")
parser.add_argument(
    "--segments-per-chunk",
    type=int,
    default=100,
    help="Number of videos per chunk to save.",
)
parser.add_argument(
    "-j",
    "--num-workers",
    type=int,
    default=4,
    help="Number of workers to run the task.",
)
parser.add_argument("--unlabelled", action="store_false", dest="labelled")


def main(args):
    if args.labels.suffix.lower() == ".pkl":
        labels = pd.read_pickle(args.labels)
    elif args.labels.suffix.lower() == ".csv":
        labels = pd.read_csv(args.labels, index_col="uid")
    else:
        raise ValueError("Expected .csv or .pkl suffix for annotation file")
    if args.modality.lower() == "flow":
        epic_adapter = adapter.EpicFlowDatasetAdapter(
            str(args.in_folder),
            labels,
            args.frame_size,
            args.extension,
            labelled=args.labelled,
        )
    elif args.modality.lower() == "rgb":
        epic_adapter = adapter.EpicDatasetAdapter(
            str(args.in_folder),
            labels,
            args.frame_size,
            args.extension,
            labelled=args.labelled,
        )
    else:
        raise ValueError("Modality '{}' not supported".format(args.modality))
    ingestor = GulpIngestor(
        epic_adapter, str(args.out_folder), args.segments_per_chunk, args.num_workers
    )
    ingestor()


if __name__ == "__main__":
    main(parser.parse_args())



#!/usr/bin/env python

import os
from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, "epic_kitchens", "__version__.py"), "r") as f:
    exec(f.read(), about)


setup(
    name=about["__title__"],
    description=about["__description__"],
    version=about["__version__"],
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "gulp_epic = epic_kitchens.gulp:main",
            "segment_epic = epic_kitchens.preprocessing.split_segments:main",
        ]
    },
    install_requires=["gulpio>=500", "pandas", "numpy"],
    extras_require={
        "dev": ["Sphinx", "pygments", "sphinx_rtd_theme", "pytest", "mypy"],
        "visualisation": ["moviepy>=1.0.0", "ipython"],
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords=["dataset", "egocentric", "action-recogntion", "epic", "epic-kitchens"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    license=about["__license__"],
    url="http://github.com/epic-kitchens/epic-lib",
    project_urls={
        "Bug Tracker": "https://github.com/epic-kitchens/epic-lib/issues",
        "Documentation": "https://epic-kitchens.readthedocs.io",
        "Dataset Website": "https://epic-kitchens.github.io/",
        "Source Code": "http://github.com/epic-kitchens/epic-lib",
    },
)



#!/usr/bin/env python3

"""Program for splitting frames into action segments
See :ref:`cli_tools_action_segmentation` for usage details """

import argparse
import logging
import os
import pathlib
import sys

import pandas as pd

from epic_kitchens.labels import VIDEO_ID_COL
from epic_kitchens.video import (
    ModalityIterator,
    FlowModalityIterator,
    RGBModalityIterator,
    split_video_frames,
)

HELP = """\
Process frame dumps, and a set of annotations in a pickled dataframe
to produce a set of segmented action videos using symbolic links.


Taking a set of videos in the directory format (for RGB):

    P01_01
    |--- frame_0000000001.jpg
    |--- frame_0000000002.jpg
    |--- ...

Produce a set of action segments in the directory format:

    P01_01_0_chop-wood
    |--- frame_0000000001.jpg
    |--- ...
    |--- frame_0000000735.jpg


The final number `Z` in `PXX_YY_Z-narration` denotes the index of the segment, this can then
be used to look up the corresponding information on the segment such as the raw narration,
verb class, noun classes etc

If segmenting optical flow then frames are contained in a `u` or `v` subdirectory.
"""

LOG = logging.getLogger(__name__)


parser = argparse.ArgumentParser(
    description=HELP, formatter_class=argparse.RawTextHelpFormatter
)
parser.add_argument("video", type=str, help="Video ID to segment")
parser.add_argument(
    "frame_dir",
    type=lambda p: pathlib.Path(p).absolute(),
    help="Path to frames, if RGB should contain images, if flow, should contain u, "
    "v subdirectories with images",
)
parser.add_argument(
    "links_dir",
    type=lambda p: pathlib.Path(p).absolute(),
    help="Path to save segments into",
)
parser.add_argument(
    "labels",
    type=pathlib.Path,
    help="Path to the pickle or CSV file which contains the meta information about the dataset.",
)
parser.add_argument(
    "modality",
    type=str.lower,
    default="rgb",
    choices=["rgb", "flow"],
    help="Modality of frames that are being segmented",
)
parser.add_argument(
    "--frame-format",
    type=str,
    default="frame_%010d.jpg",
    help="Pattern of frame filenames (default: %(default)s)",
)
parser.add_argument(
    "--fps",
    type=float,
    default=60,
    help="FPS of extracted frames (default: %(default)s)",
)
parser.add_argument(
    "--of-stride",
    type=int,
    default=2,
    help="Optical flow stride parameter used for frame extraction (default: %(default)s)",
)
parser.add_argument(
    "--of-dilation",
    type=int,
    default=3,
    help="Optical flow dilation parameter used for frame extraction "
    "(default: %(default)s)",
)


def main(args):
    logging.basicConfig(level=logging.INFO)
    if not args.labels.exists():
        LOG.error("Annotations pickle: '{}' does not exist".format(args.labels))
        sys.exit(1)

    if args.labels.suffix.lower() == ".pkl":
        annotations = pd.read_pickle(args.labels)
    elif args.labels.suffix.lower() == ".csv":
        annotations = pd.read_csv(args.labels, index_col="uid")
    else:
        raise ValueError("Expected .csv or .pkl suffix for annotation file")

    fps = float(args.fps)
    if args.modality.lower() == "rgb":
        frame_dirs = [args.frame_dir]
        links_dirs = [args.links_dir]
        modality = RGBModalityIterator(fps=fps)  # type: ModalityIterator
    elif args.modality.lower() == "flow":
        axes = ["u", "v"]
        frame_dirs = [args.frame_dir.joinpath(axis) for axis in axes]
        links_dirs = [args.links_dir.joinpath(axis) for axis in axes]
        modality = FlowModalityIterator(
            rgb_fps=fps, stride=int(args.of_stride), dilation=int(args.of_dilation)
        )
    else:
        raise ValueError("Modality '{}' is not recognised".format(args.modality))

    video_annotations = annotations[annotations[VIDEO_ID_COL] == args.video]
    for frame_dir, links_dir in zip(frame_dirs, links_dirs):
        split_video_frames(
            modality, args.frame_format, video_annotations, links_dir, frame_dir
        )


if __name__ == "__main__":
    main(parser.parse_args())



import logging
import os
import re
from abc import ABC
from pathlib import Path

import numpy as np
import pandas as pd
from typing import Tuple, Iterator, List, Iterable

from epic_kitchens.labels import NARRATION_COL, START_TS_COL, STOP_TS_COL, VIDEO_ID_COL
from epic_kitchens.time import timestamp_to_frame

LOG = logging.getLogger(__name__)

Timestamp = str


class ModalityIterator(ABC):
    """Interface that a modality extracted from video must implement"""

    def frame_iterator(self, start: Timestamp, stop: Timestamp) -> Iterable[int]:
        """
        Args:
            start: start time (timestamp: ``HH:MM:SS[.FractionalPart]``)
            stop: stop time (timestamp: ``HH:MM:SS[.FractionalPart]``)

        Yields:
            Frame indices iterator corresponding to segment from ``start`` to ``stop``
        """
        raise NotImplementedError


class RGBModalityIterator(ModalityIterator):
    """Iterator for RGB frames"""

    def __init__(self, fps):
        self.fps = fps

    def frame_iterator(self, start: Timestamp, stop: Timestamp) -> Iterable[int]:
        start_frame = timestamp_to_frame(start, self.fps)
        stop_frame = timestamp_to_frame(stop, self.fps)
        return range(start_frame, stop_frame)


class FlowModalityIterator(ModalityIterator):
    """Iterator for optical flow :math:`(u, v)` frames"""

    def __init__(self, dilation=1, stride=1, bound=20, rgb_fps=59.94):
        """
        Args:
            dilation: Dilation that optical flow was extracted with
            stride: Stride that optical flow was extracted with
            bound: Bound that optical flow was extracted with
            rgb_fps: FPS of RGB video flow was computed from
        """
        self.dilation = dilation
        self.stride = stride
        self.bound = bound
        self.rgb_fps = rgb_fps

    def frame_iterator(self, start: Timestamp, stop: Timestamp) -> Iterable[int]:
        start_frame = self._seconds_to_flow_frame_index(start)
        stop_frame = self._seconds_to_flow_frame_index(stop)
        return range(start_frame, stop_frame)

    def _seconds_to_flow_frame_index(self, timestamp: str):
        rgb_frame_index = timestamp_to_frame(timestamp, self.rgb_fps)
        return int(np.ceil(rgb_frame_index / self.stride))


def iterate_frame_dir(root: Path) -> Iterator[Tuple[Path, Path]]:
    """ Iterate over a directory of video dirs with the hierarchy ``root/P01/P01_01/``

    Args:
        root: Root directory with person directory children, then each person directory has
              video directory children e.g. root -> P01 -> P01_01
    Yields:
        (person_dir, video_dir)
    """
    for person_dir in root.iterdir():
        for video_dir in person_dir.iterdir():
            yield (person_dir, video_dir)


def split_dataset_frames(
    modality_iterator: ModalityIterator,
    frames_dir: Path,
    segment_root_dir: Path,
    annotations: pd.DataFrame,
    frame_format="frame%06d.jpg",
    pattern=re.compile(".*"),
) -> None:
    """Split dumped video frames from ``frames_dir`` into directories within ``segment_root_dir`` for each
    video segment defined in ``annotations``.

    Args:
        modality_iterator: Modality iterator of frames
        frames_dir: Directory containing dumped frames
        segment_root_dir: Directory to write split segments to
        annotations: Dataframe containing segment information
        frame_format (str, optional): Old style string format that must contain a single ``%d`` formatter
            describing file name format of the dumped frames.
        pattern (re.Pattern, optional): Regexp to match video directories
    """
    assert frames_dir.exists()

    frames_dir = frames_dir.resolve()
    segment_root_dir.mkdir(exist_ok=True, parents=True)
    segment_root_dir = segment_root_dir.resolve()

    for person_dir, video_dir in iterate_frame_dir(frames_dir):
        if pattern.search(str(video_dir)):
            annotations_for_video = annotations[
                annotations[VIDEO_ID_COL] == video_dir.name
            ]
            split_video_frames(
                modality_iterator,
                frame_format,
                annotations_for_video,
                segment_root_dir,
                video_dir,
            )


def get_narration(annotation):
    """Get narration from annotation row, defaults to ``"unnarrated"`` if row has no narration column."""
    try:
        return getattr(annotation, NARRATION_COL)
    except AttributeError:
        return "unnarrated"


def split_video_frames(
    modality_iterator: ModalityIterator,
    frame_format: str,
    video_annotations: pd.DataFrame,
    segment_root_dir: Path,
    video_dir: Path,
) -> None:
    """Split frames from a single video file stored in ``video_dir`` into segment directories stored
    in ``segment_root_dir``.

    Args:
        modality_iterator: Modality iterator
        frame_format: Old style string format that must contain a single ``%d`` formatter
            describing file name format of the dumped frames.
        video_annotations: Dataframe containing rows only corresponding to video frames stored in
            :param:`video_dir`
        segment_root_dir: Directory to write split segments to
        video_dir: Directory containing dumped frames for a single video
    """
    for annotation in video_annotations.itertuples():
        segment_dir_name = "{video_id}_{index}_{narration}".format(
            index=annotation.Index,
            video_id=getattr(annotation, VIDEO_ID_COL),
            narration=get_narration(annotation).strip().lower().replace(" ", "-"),
        )
        segment_dir = segment_root_dir.joinpath(segment_dir_name)
        segment_dir.mkdir(parents=True, exist_ok=True)
        start_timestamp = getattr(annotation, START_TS_COL)
        stop_timestamp = getattr(annotation, STOP_TS_COL)
        frame_iterator = modality_iterator.frame_iterator(
            start_timestamp, stop_timestamp
        )

        LOG.info(
            "Linking {video_id} - {narration} - {start}--{stop}".format(
                video_id=getattr(annotation, VIDEO_ID_COL),
                narration=get_narration(annotation),
                start=start_timestamp,
                stop=stop_timestamp,
            )
        )
        _split_frames_by_segment(frame_format, frame_iterator, segment_dir, video_dir)


def _split_frames_by_segment(
    frame_format: str, frame_iterator, segment_dir: Path, video_dir: Path
):
    segment_dir_fd = os.open(str(segment_dir), os.O_RDONLY)
    try:
        first_frame_index = -1
        last_frame_index = -1
        for frame_index in frame_iterator:
            if first_frame_index == -1:
                first_frame_index = frame_index
            source_frame_filename = frame_format % frame_index
            target_frame_filename = frame_format % (frame_index - first_frame_index + 1)
            source_frame_path = video_dir.joinpath(source_frame_filename)
            segmented_frame_path = segment_dir.joinpath(target_frame_filename)
            assert (
                source_frame_path.exists()
            ), "{source_frame_path} does not exist".format(
                source_frame_path=source_frame_path
            )
            if os.path.lexists(str(segmented_frame_path)):
                os.remove(str(segmented_frame_path))
            source_frame_relative_path = os.path.relpath(
                str(source_frame_path), start=str(segment_dir)
            )
            os.symlink(
                str(source_frame_relative_path),
                str(segmented_frame_path),
                dir_fd=segment_dir_fd,
            )
            last_frame_index = frame_index
        LOG.info(
            "  Linked [{first},{last}] -> [1, {last_target}]".format(
                first=first_frame_index,
                last=last_frame_index,
                last_target=last_frame_index - first_frame_index + 1,
            )
        )
    finally:
        os.close(segment_dir_fd)



""" ``epic-kitchens``

This library contains a variety of useful classes and functions for performing common operations on
the `EPIC Kitchens`_ dataset.

.. _`EPIC Kitchens`: https://epic-kitchens.github.io/

"""

from . import gulp, dataset, preprocessing
from .__version__ import __title__, __description__, __version__
from .__version__ import __author__, __author_email__, __license__, __copyright__



from datetime import datetime

import epic_kitchens.meta
import pytest


@pytest.fixture(scope="module")
def descriptions(tmp_datadir):
    return epic_kitchens.meta.video_descriptions()


def test_descriptions_has_at_least_432_entries(descriptions):
    assert len(descriptions) >= 432


def test_descriptions_has_video_id_index(descriptions):
    assert descriptions.index.name == "video_id"


@pytest.mark.parametrize("col", ["datetime", "description"])
def test_description_has_columns(col, descriptions):
    assert col in descriptions.columns


def test_descriptions_reads_date_and_time_to_datetime(descriptions):
    assert isinstance(descriptions["datetime"].iloc[0], datetime)
